from Crypto.Cipher import AES
from Crypto.Util.number import long_to_bytes
import hashlib

from random import randint

def sha256(my_string):
	m = hashlib.sha256(my_string).digest()
	v = long_to_bytes(ZZ(m.encode('hex'), 16))
	return v

def generate_random_vector_mod(q, dimension):
	return vector([   ZZ(randint(0, q - 1)) for i in range(dimension)])

def reduction_mod(r, q):
	return (r/q - floor(r/q))*q

def vector_to_polynomial(vector):
	poly = 0
	for i in range(vector.length()):
		poly = poly + x^i * vector[i]
	return poly

def coeff_to_canonical(polynomial, vector):
	eval = []
	for i in range(len(vector)):
		eval.append(polynomial(vector[i]))
	return eval

def generate_secret(prime,degree):
	secret_coeff = vector_to_polynomial(generate_random_vector_mod(prime, degree))
	return secret_coeff


def generate_error(sigma, prime, degree):
	D = RealDistribution('gaussian',sigma)
	error = []
	for i in range(degree):
		error.append(D.get_random_element())
	error_canonical_interm = U * (vector(error).column())
	error_canonical = [(error_canonical_interm)[i][0] for i in range(degree)]
	error_coeff = []
	error_coeff_interm = Sigma_Inv * error_canonical_interm
	for i in range(degree):
		error_coeff.append(error_coeff_interm[i][0].real())
	error_coeff = vector_to_polynomial(vector(error_coeff))
	return error_coeff



def create_samples(nr_samples, prime, sigma,degree):
	print "Generating secret key"
	sec_coeff = generate_secret(prime, degree)
	sec_coeff_list = vector(sec_coeff.list())

	challenge_file=open("out/challenge.txt","w+")
	secret_file = open("priv/secret.txt","w+")

	secret_file.write('The secret is '+str(sec_coeff_list)+'\n')
	key = sha256(str(sec_coeff_list))
	iv = b'\x00' * 16
	input_file = open("in/input.jpg")
	input_data = input_file.read()
	input_file.close()

	print "Encrypting a file with the secret key"
	cfb_cipher = AES.new(key, AES.MODE_CFB, iv)
	enc_data = cfb_cipher.encrypt(input_data)

	enc_file = open("out/encrypted.enc", "w")
	enc_file.write(enc_data)
	enc_file.close()

	print "Done! Now writing the samples for you to crack"
	challenge_file.write('I shall give you some samples. Can you recover the secret key?\n')
	for i in range(nr_samples):

		a = generate_random_vector_mod(prime,degree)
		a_coeff = vector_to_polynomial(a)
		a_canonical = coeff_to_canonical(a_coeff,roots_of_defining_polynomial)
		challenge_file.write('a_'+str(i)+' is '+str(a_canonical)+'\n')
		err_coeff = generate_error(sigma, prime, degree)
		b_coeff_interm = N(a_coeff * sec_coeff + err_coeff)
		b_coeff = 0
		for j in range(degree):
			b_coeff = b_coeff + x^j*reduction_mod(b_coeff_interm.list()[j],prime)
		b_canonical = coeff_to_canonical(b_coeff, roots_of_defining_polynomial)
		challenge_file.write('b_'+str(i)+' is '+str(b_canonical)+'\n')


	challenge_file.close()
	secret_file.close()




print "Please wait while all the quantum bits warm up..."

S.<x>=PolynomialRing(RationalField(100))
prime = 1487
sigma = 1000
nr_samples = 8

defining_polynomial = x^256 + prime - 1
J = S.ideal(defining_polynomial)
N = S.quotient_ring(J)
degree = defining_polynomial.degree()
K.<alpha> = NumberField(defining_polynomial)
r, s = K.signature()

print "Nearly there..."

roots_of_defining_polynomial = []
aux = [i[0] for i in defining_polynomial.roots(CC)]
for i in range(r):
	roots_of_defining_polynomial.append(aux[i])
for i in range(s):
	roots_of_defining_polynomial.append(aux[r + i*2])
for i in range(s):
	roots_of_defining_polynomial.append(aux[r + i*2 + 1])

I = CC.0
I_r = identity_matrix(r)
I_s = identity_matrix(s)
O_r_s = zero_matrix(r,s)
O_s_r = zero_matrix(s,r)
U = block_matrix([[I_r, O_r_s, O_r_s], [O_s_r, 1/float(sqrt(2))*I_s, I/float(sqrt(2))*I_s], [O_s_r, 1/float(sqrt(2))*I_s, -I/float(sqrt(2))*I_s]], subdivide = False)

Sigma = matrix(CDF.to_prec(230), degree, degree, lambda i,j: roots_of_defining_polynomial[i]^j)
print "One more step...."
Sigma_Inv = Sigma.inverse()

create_samples(nr_samples,prime,sigma,degree)

